@@ -1,4 +1,4 @@
-#!/bin/sh
+#! /bin/sh
 
 if [ -z "$ANDROID_NDK_HOME" ]; then
     echo "You should probably set ANDROID_NDK_HOME to the directory containing"
@@ -11,7 +11,7 @@ if [ ! -f ./configure ]; then
 	exit 1
 fi
 
-if [ "x$TARGET_ARCH" = 'x' ] || [ "x$HOST_COMPILER" = 'x' ]; then
+if [ "x$TARGET_ARCH" = 'x' ] || [ "x$ARCH" = 'x' ] || [ "x$HOST_COMPILER" = 'x' ]; then
     echo "You shouldn't use android-build.sh directly, use android-[arch].sh instead"
     exit 1
 fi
@@ -22,11 +22,10 @@ export PREFIX="$(pwd)/libsodium-android-${TARGET_ARCH}"
 export TOOLCHAIN_DIR="$(pwd)/android-toolchain-${TARGET_ARCH}"
 export PATH="${PATH}:${TOOLCHAIN_DIR}/bin"
 
-# Clean up before build
 rm -rf "${TOOLCHAIN_DIR}" "${PREFIX}"
 
-$MAKE_TOOLCHAIN --platform="${NDK_PLATFORM:-android-14}" \
-                --arch="$TARGET_ARCH" \
+bash $MAKE_TOOLCHAIN --platform="${NDK_PLATFORM:-android-21}" \
+                --arch="$ARCH" \
                 --install-dir="$TOOLCHAIN_DIR" && \
 ./configure --host="${HOST_COMPILER}" \
             --with-sysroot="${TOOLCHAIN_DIR}/sysroot" \
