@@ -4,6 +4,9 @@
 # include <sys/stat.h>
 # include <sys/time.h>
 #endif
+#ifdef __linux__
+# include <sys/syscall.h>
+#endif
 
 #include <assert.h>
 #include <errno.h>
@@ -43,12 +46,6 @@ randombytes_sysrandom_stir(void)
 {
 }
 
-uint32_t
-randombytes_sysrandom_uniform(const uint32_t upper_bound)
-{
-    return arc4random_uniform(upper_bound);
-}
-
 void
 randombytes_sysrandom_buf(void * const buf, const size_t size)
 {
@@ -64,35 +61,37 @@ randombytes_sysrandom_close(void)
 #else /* __OpenBSD__ */
 
 typedef struct SysRandom_ {
-    int        random_data_source_fd;
-    int        initialized;
+    int random_data_source_fd;
+    int initialized;
+    int getrandom_available;
 } SysRandom;
 
 static SysRandom stream = {
     SODIUM_C99(.random_data_source_fd =) -1,
-    SODIUM_C99(.initialized =) 0
+    SODIUM_C99(.initialized =) 0,
+    SODIUM_C99(.getrandom_available =) 0
 };
 
 #ifndef _WIN32
 static ssize_t
-safe_read(const int fd, void * const buf_, size_t count)
+safe_read(const int fd, void * const buf_, size_t size)
 {
     unsigned char *buf = (unsigned char *) buf_;
     ssize_t        readnb;
 
-    assert(count > (size_t) 0U);
+    assert(size > (size_t) 0U);
     do {
-        while ((readnb = read(fd, buf, count)) < (ssize_t) 0 &&
-               errno == EINTR); /* LCOV_EXCL_LINE */
+        while ((readnb = read(fd, buf, size)) < (ssize_t) 0 &&
+               (errno == EINTR || errno == EAGAIN)); /* LCOV_EXCL_LINE */
         if (readnb < (ssize_t) 0) {
             return readnb; /* LCOV_EXCL_LINE */
         }
         if (readnb == (ssize_t) 0) {
             break; /* LCOV_EXCL_LINE */
         }
-        count -= (size_t) readnb;
+        size -= (size_t) readnb;
         buf += readnb;
-    } while (count > (ssize_t) 0);
+    } while (size > (ssize_t) 0);
 
     return (ssize_t) (buf - (unsigned char *) buf_);
 }
@@ -114,23 +113,79 @@ randombytes_sysrandom_random_dev_open(void)
     int                fd;
 
     do {
-        if ((fd = open(*device, O_RDONLY)) != -1) {
+        fd = open(*device, O_RDONLY);
+        if (fd != -1) {
             if (fstat(fd, &st) == 0 && S_ISCHR(st.st_mode)) {
+# if defined(F_SETFD) && defined(FD_CLOEXEC)
+                (void) fcntl(fd, F_SETFD, fcntl(fd, F_GETFD) | FD_CLOEXEC);
+# endif
                 return fd;
             }
             (void) close(fd);
+        } else if (errno == EINTR) {
+            continue;
         }
         device++;
     } while (*device != NULL);
 
+    errno = EIO;
     return -1;
 /* LCOV_EXCL_STOP */
 }
 
+# ifdef SYS_getrandom
+static int
+_randombytes_linux_getrandom(void * const buf, const size_t size)
+{
+    int readnb;
+
+    assert(size <= 256U);
+    do {
+        readnb = syscall(SYS_getrandom, buf, (int) size, 0);
+    } while (readnb < 0 && (errno == EINTR || errno == EAGAIN));
+
+    return (readnb == (int) size) - 1;
+}
+
+static int
+randombytes_linux_getrandom(void * const buf_, size_t size)
+{
+    unsigned char *buf = (unsigned char *) buf_;
+    size_t         chunk_size = 256U;
+
+    do {
+        if (size < chunk_size) {
+            chunk_size = size;
+            assert(chunk_size > (size_t) 0U);
+        }
+        if (_randombytes_linux_getrandom(buf, chunk_size) != 0) {
+            return -1;
+        }
+        size -= chunk_size;
+        buf += chunk_size;
+    } while (size > (size_t) 0U);
+
+    return 0;
+}
+# endif
+
 static void
 randombytes_sysrandom_init(void)
 {
-    const int errno_save = errno;
+    const int     errno_save = errno;
+
+# ifdef SYS_getrandom
+    {
+	unsigned char fodder[16];
+
+	if (randombytes_linux_getrandom(fodder, sizeof fodder) == 0) {
+	    stream.getrandom_available = 1;
+	    errno = errno_save;
+	    return;
+	}
+	stream.getrandom_available = 0;
+    }
+# endif
 
     if ((stream.random_data_source_fd =
          randombytes_sysrandom_random_dev_open()) == -1) {
@@ -176,6 +231,11 @@ randombytes_sysrandom_close(void)
         stream.initialized = 0;
         ret = 0;
     }
+# ifdef SYS_getrandom
+    if (stream.getrandom_available != 0) {
+        ret = 0;
+    }
+# endif
 #else /* _WIN32 */
     if (stream.initialized != 0) {
         stream.initialized = 0;
@@ -204,7 +264,16 @@ randombytes_sysrandom_buf(void * const buf, const size_t size)
     assert(size <= ULONG_LONG_MAX);
 #endif
 #ifndef _WIN32
-    if (safe_read(stream.random_data_source_fd, buf, size) != (ssize_t) size) {
+# ifdef SYS_getrandom
+    if (stream.getrandom_available != 0) {
+        if (randombytes_linux_getrandom(buf, size) != 0) {
+            abort();
+        }
+        return;
+    }
+# endif
+    if (stream.random_data_source_fd == -1 ||
+        safe_read(stream.random_data_source_fd, buf, size) != (ssize_t) size) {
         abort(); /* LCOV_EXCL_LINE */
     }
 #else
@@ -217,31 +286,7 @@ randombytes_sysrandom_buf(void * const buf, const size_t size)
 #endif
 }
 
-/*
- * randombytes_sysrandom_uniform() derives from OpenBSD's arc4random_uniform()
- * Copyright (c) 2008, Damien Miller <djm@openbsd.org>
- */
-
-uint32_t
-randombytes_sysrandom_uniform(const uint32_t upper_bound)
-{
-    uint32_t min;
-    uint32_t r;
-
-    if (upper_bound < 2) {
-        return 0;
-    }
-    min = (uint32_t) (-upper_bound % upper_bound);
-    for (;;) {
-        r = randombytes_sysrandom();
-        if (r >= min) {
-            break;
-        }
-    } /* LCOV_EXCL_LINE */
-    return r % upper_bound;
-}
-
-#endif
+#endif /* __OpenBSD__ */
 
 const char *
 randombytes_sysrandom_implementation_name(void)
@@ -253,7 +298,7 @@ struct randombytes_implementation randombytes_sysrandom_implementation = {
     SODIUM_C99(.implementation_name =) randombytes_sysrandom_implementation_name,
     SODIUM_C99(.random =) randombytes_sysrandom,
     SODIUM_C99(.stir =) randombytes_sysrandom_stir,
-    SODIUM_C99(.uniform =) randombytes_sysrandom_uniform,
+    SODIUM_C99(.uniform =) NULL,
     SODIUM_C99(.buf =) randombytes_sysrandom_buf,
     SODIUM_C99(.close =) randombytes_sysrandom_close
 };
